/*!
        @file    $Id:: field.cpp #$

        @brief

        @author  <Hideo Matsufuru> hideo.matsufuru@kek.jp(matsufuru) 
                 $LastChangedBy: namekawa $

        @date    $LastChangedDate:: 2013-06-19 01:18:08 #$

        @version $LastChangedRevision: 897 $
*/

#include "field.h"
#include <fstream>

#include "threadManager_OpenMP.h"

#include "bridgeIO.h"
using Bridge::vout;

using std::string;

//====================================================================
void Field::setpart_ex(int ex, const Field& w, int exw){

  assert(ex < m_Nex);
  assert(exw < w.nex());

  int nth = ThreadManager_OpenMP::get_num_threads();
  int ith = ThreadManager_OpenMP::get_thread_id();

  int ndof = m_Nin * m_Nvol;
  int is  = ndof * ith / nth;
  int ns  = ndof * (ith+1) / nth - is;

  double* yp = ptr(is + ndof*ex);
  double* wp = const_cast<Field*>(&w)->ptr(is + ndof*exw);

  for(int i = 0; i < ns; ++i){
    yp[i] = wp[i];
  }

}

//====================================================================
void Field::addpart_ex(int ex, const Field& w, int exw){

  assert(ex < m_Nex);
  assert(exw < w.nex());

  int nth = ThreadManager_OpenMP::get_num_threads();
  int ith = ThreadManager_OpenMP::get_thread_id();

  int ndof = m_Nin * m_Nvol;
  int is  = ndof * ith / nth;
  int ns  = ndof * (ith+1) / nth - is;

  double* yp = ptr(is + ndof*ex);
  double* wp = const_cast<Field*>(&w)->ptr(is + ndof*exw);

  for(int i = 0; i < ns; ++i){
    yp[i] += wp[i];
  }

}

//====================================================================
void Field::addpart_ex(int ex, const Field& w, int exw, double prf){

  assert(ex < m_Nex);
  assert(exw < w.nex());

  int nth = ThreadManager_OpenMP::get_num_threads();
  int ith = ThreadManager_OpenMP::get_thread_id();

  int ndof = m_Nin * m_Nvol;
  int is  = ndof * ith / nth;
  int ns  = ndof * (ith+1) / nth - is;

  double* yp = ptr(is + ndof*ex);
  double* wp = const_cast<Field*>(&w)->ptr(is + ndof*exw);

  for(int i = 0; i < ns; ++i){
    yp[i] += prf * wp[i];
  }

}

//====================================================================
double Field::ddotc_ex(const int ex1, const Field& x, const int ex2)
 const {

  int nth = ThreadManager_OpenMP::get_num_threads();
  int ith = ThreadManager_OpenMP::get_thread_id();

  int Niv = m_Nin * m_Nvol;
  int is  = Niv * ith / nth;
  int ns  = Niv *(ith+1)/nth - is;

  double* yp = const_cast<Field*>(this)->ptr(is + Niv*ex1);
  double* xp = const_cast<Field*>(&x)->ptr(is + Niv*ex2);

  double a = 0.0;

  for(int i = 0; i < ns; ++i){
    a += yp[i] * xp[i];
  }

  ThreadManager_OpenMP::reduce_sum_global(a,ith,nth);
  return a;

}

//====================================================================
double Field::ddotc(const Field& x) const {

  int nth = ThreadManager_OpenMP::get_num_threads();
  int ith = ThreadManager_OpenMP::get_thread_id();

  int is  = m_Ntot * ith / nth;
  int ns  = m_Ntot*(ith+1)/nth - is;

  double* yp = const_cast<Field*>(this)->ptr(is);
  double* xp = const_cast<Field*>(&x)->ptr(is);

  double a = 0.0;

  for(int i = 0; i < ns; ++i){
    a += yp[i] * xp[i];
  }

  ThreadManager_OpenMP::reduce_sum_global(a,ith,nth);
  return a;

}

//====================================================================
void Field::ddotc(double& yx, const Field& x){

  int nth = ThreadManager_OpenMP::get_num_threads();
  int ith = ThreadManager_OpenMP::get_thread_id();

  int is  = m_Ntot * ith / nth;
  int ns  = m_Ntot*(ith+1)/nth - is;

  double* yp = const_cast<Field*>(this)->ptr(is);
  double* xp = const_cast<Field*>(&x)->ptr(is);

  double a = 0.0;

  for(int i = 0; i < ns; ++i){
    a += yp[i] * xp[i];
  }

  ThreadManager_OpenMP::reduce_sum_global(a,ith,nth);
  yx = a;

}

//====================================================================
dcomplex Field::zdotc(const Field& x) const {

  if(field_complexness() == Field::COMPLEX){

    return zdotc_cmplx(x);

  }else if(field_complexness()==Field::REAL){

    double prdr = ddotc(x);
    return cmplx(prdr,0.0);

  }else{

    abort();
    return cmplx(0.0,0.0);

  }

}

//====================================================================
dcomplex Field::zdotc_cmplx(const Field& x) const {

  //  int Ntot = ntot();
  int Ntot2 = m_Ntot/2;
  assert(x.ntot() == m_Ntot);

  int nth = ThreadManager_OpenMP::get_num_threads();
  int ith = ThreadManager_OpenMP::get_thread_id();

  int is  = Ntot2 * ith / nth;
  int ns  = Ntot2*(ith+1)/nth - is;

  double* yp = const_cast<Field*>(this)->ptr(2*is);
  double* xp = const_cast<Field*>(&x)->ptr(2*is);

  double prdr = 0.0;
  double prdi = 0.0;

  for(int k = 0; k < ns; ++k){
      prdr += yp[2*k  ] * xp[2*k  ]
             +yp[2*k+1] * xp[2*k+1];
      prdi += yp[2*k  ] * xp[2*k+1]
             -yp[2*k+1] * xp[2*k  ];
  };

  prdr = Communicator::reduce_sum(prdr);
  prdi = Communicator::reduce_sum(prdi);

  return cmplx(prdr,prdi);

}

//====================================================================
double Field::norm2() const {

  int nth = ThreadManager_OpenMP::get_num_threads();
  int ith = ThreadManager_OpenMP::get_thread_id();

  int is  = m_Ntot * ith / nth;
  int ns  = m_Ntot*(ith+1)/nth - is;

  double* yp = const_cast<Field*>(this)->ptr(is);

  double a = 0.0;
  for(int i = 0; i < ns; ++i){
    a += yp[i] * yp[i];
  }

  ThreadManager_OpenMP::reduce_sum_global(a,ith,nth);
  return a;

}

//====================================================================
void Field::norm2(double& yy){

  int nth = ThreadManager_OpenMP::get_num_threads();
  int ith = ThreadManager_OpenMP::get_thread_id();

  int is  = m_Ntot * ith / nth;
  int ns  = m_Ntot*(ith+1)/nth - is;

  double* yp = ptr(is);

  double a = 0.0;
  for(int i = 0; i < ns; ++i){
    a += yp[i] * yp[i];
  }

  ThreadManager_OpenMP::reduce_sum_global(a,ith,nth);
  yy = a;

}

//====================================================================
void Field::daxpy(double a, const Field& x){

  int nth = ThreadManager_OpenMP::get_num_threads();
  int ith = ThreadManager_OpenMP::get_thread_id();

  int is  = m_Ntot * ith / nth;
  int ns  = m_Ntot*(ith+1)/nth - is;

  double* yp = ptr(is);
  double* xp = const_cast<Field*>(&x)->ptr(is);

  for(int i = 0; i < ns; ++i){
    yp[i] += a * xp[i];
  }

}

//====================================================================
void Field::zaxpy(dcomplex a, const Field& x){

  if(m_complexness == COMPLEX){

    zaxpy(a,x);

  }else if(m_complexness == REAL){

    double ar = real(a);
    daxpy(ar,x);

  }else{
    abort();
  }

}

//====================================================================
void Field::zaxpy_cmplx(dcomplex a, const Field& x){

  int Ntot2 = m_Ntot/2;
  assert(x.ntot() == m_Ntot);

  int nth = ThreadManager_OpenMP::get_num_threads();
  int ith = ThreadManager_OpenMP::get_thread_id();

  int is  = Ntot2 * ith / nth;
  int ns  = Ntot2*(ith+1)/nth - is;

  double* yp = const_cast<Field*>(this)->ptr(2*is);
  double* xp = const_cast<Field*>(&x)->ptr(2*is);

  double ar = real(a);
  double ai = imag(a);
  for(int k = 0; k < ns; ++k){
    yp[2*k  ] +=  ar * xp[2*k] - ai * xp[2*k+1];
    yp[2*k+1] +=  ai * xp[2*k] + ar * xp[2*k+1];
  }

}

//====================================================================
void Field::zaxpy_copy(dcomplex a, const Field& x, const Field& w){

  if(m_complexness == COMPLEX){

    zaxpy_copy_cmplx(a,x,w);

  }else if(m_complexness == REAL){

    double ar = real(a);
    daxpy_copy(ar,x,w);

  }else{
    abort();
  }

}

//====================================================================
void Field::zaxpy_copy_cmplx(dcomplex a, const Field& x, const Field& w)
{

  int Ntot2 = m_Ntot/2;
  assert(x.ntot() == m_Ntot);

  int nth = ThreadManager_OpenMP::get_num_threads();
  int ith = ThreadManager_OpenMP::get_thread_id();

  int is  = Ntot2 * ith / nth;
  int ns  = Ntot2*(ith+1)/nth - is;

  double* yp = const_cast<Field*>(this)->ptr(2*is);
  double* xp = const_cast<Field*>(&x)->ptr(2*is);
  double* wp = const_cast<Field*>(&w)->ptr(2*is);

  double ar = real(a);
  double ai = imag(a);
  for(int k = 0; k < ns; ++k){
    yp[2*k  ] = wp[2*k  ] + ar * xp[2*k] - ai * xp[2*k+1];
    yp[2*k+1] = wp[2*k+1] + ai * xp[2*k] + ar * xp[2*k+1];
  }

}

//====================================================================
void Field::daxpy_copy(double a, const Field& x, const Field& w){

  int nth = ThreadManager_OpenMP::get_num_threads();
  int ith = ThreadManager_OpenMP::get_thread_id();

  int is  = m_Ntot * ith / nth;
  int ns  = m_Ntot*(ith+1)/nth - is;

  double* yp = ptr(is);
  double* xp = const_cast<Field*>(&x)->ptr(is);
  double* wp = const_cast<Field*>(&w)->ptr(is);

  for(int i = 0; i < ns; ++i){
    yp[i] = wp[i] + a * xp[i];
  }

}


//====================================================================
void Field::dscal(double a) {

  int nth = ThreadManager_OpenMP::get_num_threads();
  int ith = ThreadManager_OpenMP::get_thread_id();

  int is  = m_Ntot * ith / nth;
  int ns  = m_Ntot*(ith+1)/nth - is;

  double* yp = ptr(is);

  for(int i = 0; i < ns; ++i){
    yp[i] *= a;
  }
}

//====================================================================
void Field::dcopy(const Field& x){

  int nth = ThreadManager_OpenMP::get_num_threads();
  int ith = ThreadManager_OpenMP::get_thread_id();

  int is  = m_Ntot * ith / nth;
  int ns  = m_Ntot*(ith+1)/nth - is;

  double* yp = ptr(is);
  double* xp = const_cast<Field*>(&x)->ptr(is);

  for(int k = 0; k < ns; ++k){
    yp[k] = xp[k];
  };

}

//====================================================================
void Field::dcopy(double a, const Field& x){

  int nth = ThreadManager_OpenMP::get_num_threads();
  int ith = ThreadManager_OpenMP::get_thread_id();

  int is  = m_Ntot * ith / nth;
  int ns  = m_Ntot*(ith+1)/nth - is;

  double* yp = ptr(is);
  double* xp = const_cast<Field*>(&x)->ptr(is);

  for(int k = 0; k < ns; ++k){
    yp[k] = a * xp[k];
  };

}

//====================================================================
void Field::clear(){

  int nth = ThreadManager_OpenMP::get_num_threads();
  int ith = ThreadManager_OpenMP::get_thread_id();

  int is  = m_Ntot * ith / nth;
  int ns  = m_Ntot*(ith+1)/nth - is;

  double* yp = ptr(is);

  for(int i = 0; i < ns; ++i){
    yp[i] = 0.0;
  }

}

//====================================================================
//============================================================END=====
