/*!
        @file    $Id:: solver_CG.cpp #$

        @brief

        @author  <Hideo Matsufuru> hideo.matsufuru@kek.jp(matsufuru)
                 $LastChangedBy: matufuru $

        @date    $LastChangedDate:: 2014-02-04 23:44:44 #$

        @version $LastChangedRevision: 994 $
*/

#include "solver_CG.h"

#include "threadManager_OpenMP.h"

const std::string Solver_CG::class_name = "Solver_CG";

//====================================================================
void Solver_CG::set_parameters(const int Niter, const double Stop_cond)
{
  //- print input parameters
  vout.general(m_vl, "Parameters of Solver_CG:\n");
  vout.general(m_vl, "  Niter     = %d\n", Niter);
  vout.general(m_vl, "  Stop_cond = %16.8e\n", Stop_cond);

  //- store values
  m_Niter     = Niter;
  m_Stop_cond = Stop_cond;
}


//====================================================================
void Solver_CG::solve(Field& xq, const Field& b,
                      int& Nconv, double& diff)
{
  vout.detailed(m_vl, "  CG solver starts\n");

  reset_field(b);

  vout.paranoiac(m_vl, "    norm of b = %16.8e\n", b.norm2());
  vout.paranoiac(m_vl, "    size of b = %d\n", b.size());

  double snorm = 1.0 / b.norm2();
  double rr;

  Nconv = -1;
  s     = b;

  solve_init(b, rr);

  vout.detailed(m_vl, "    iter: %8d  %22.15e\n", 0, rr * snorm);

  for (int iter = 0; iter < m_Niter; iter++) {
    solve_step(rr);

    vout.detailed(m_vl, "    iter: %8d  %22.15e\n", (iter + 1), rr * snorm);

    if (rr * snorm < m_Stop_cond) {

     #pragma omp parallel
     {
      m_fopr->mult(s, x);
      s.daxpy(-1.0, b);    //   s   -= b;
      double diff2;        //   diff = s.norm();
      s.norm2(diff2);
      #pragma omp master
        diff = sqrt(diff2);
     }

      if (diff * diff * snorm < m_Stop_cond) {
        Nconv = iter + 1;
        break;
      }

      s.dcopy(x);   // s = x;
      solve_init(b, rr);
    }
  }
  if (Nconv == -1) {
    vout.crucial(m_vl, "CG solver not converged.\n");
    abort();
  }

#pragma omp parallel
 {
  xq.dcopy(x);       //  xq = x;
 }

}

//====================================================================
void Solver_CG::reset_field(const Field& b)
{
  int Nin  = b.nin();
  int Nvol = b.nvol();
  int Nex  = b.nex();

  if ((s.nin() != Nin) || (s.nvol() != Nvol) || (s.nex() != Nex)) {

    s.reset(Nin, Nvol, Nex);
    r.reset(Nin, Nvol, Nex);
    x.reset(Nin, Nvol, Nex);
    p.reset(Nin, Nvol, Nex);
    v.reset(Nin, Nvol, Nex);

    vout.paranoiac(m_vl, "  Solver_CG: field size reset.\n");
  }
}

//====================================================================
void Solver_CG::solve_init(const Field& b, double& rr)
{

#pragma omp parallel
 {
   double rr2;

   r.dcopy(b);      //  r = b;
   x.dcopy(s);      //  x = s;

  m_fopr->mult(s, x);

  r.daxpy(-1.0,s);  //  r -= s;
  p.dcopy(r);       //  p  = r;
  r.norm2(rr2);     //  rr = r * r;

#pragma omp master
  rr = rr2;
 }

}
//====================================================================
void Solver_CG::solve_step(double& rr)
{

#pragma omp parallel
 {
  m_fopr->mult(s, p);

  double pap, rr_p, cr, rr2;

  p.ddotc(pap, s);   //  pap  = p * s;

  rr_p = rr;
  cr   = rr_p / pap;

  x.daxpy(cr,p);     //  x += cr*p;

  r.daxpy(-cr, s);   //  s *= cr;  r -= s;

  r.norm2(rr2);      //  rr = r * r;

  p.dscal(rr2/rr_p); //  p *= rr / rr_p;
  p.daxpy(1.0,r);    //  p += r;

#pragma omp master
  rr = rr2;
 }

}
//====================================================================
//============================================================END=====
