/*!
        @file    $Id:: test_Spectrum_Domainwall_2ptFunction.cpp #$

        @brief

        @author  Hideo Matsufuru <hideo.matsufuru@kek.jp> (matsufuru)
                 $LastChangedBy: matufuru $

        @date    $LastChangedDate:: 2014-06-03 23:17:04 #$

        @version $LastChangedRevision: 1057 $
*/

//#include "test.h"

#include "timer.h"

#include "gaugeConfig.h"
#include "staples.h"

#include "fopr_Wilson.h"
#include "fopr_Domainwall.h"
#include "fprop_Domainwall_4d.h"

#include "solver_CG.h"

#include "gammaMatrixSet_Chiral.h"
#include "source.h"
#include "source_Local.h"

#include "corr2pt_4spinor.h"

//====================================================================
//! Test of spectroscopy for the domain-wall fermion operator.

/*!
    This class determine the domain-wall fermion propagator and
    construct hadron correlator.
                                          [12 Dec 2012 H.Matsufuru]
    (Coding history will be recovered from trac.)
    YAML is implemented.                  [14 Nov 2012 Y.Namekawa]
    Selectors are implemented.            [03 Mar 2013 Y.Namekawa]
    (Selectors are replaced with factories by Aoyama-san)
    Largely simplified for benchmark.     [05 Jun 2014 H.Matsufuru]
 */

namespace Test_Spectrum_Domainwall {


  //====================================================================
  int hadron_2ptFunction(void)
  {
    // ####  parameter setup  ####
    int Nc   = CommonParameters::Nc();
    int Nd   = CommonParameters::Nd();
    int Nvol = CommonParameters::Nvol();
    int Lvol = CommonParameters::Lvol();
    int Ndim = CommonParameters::Ndim();

    // set verbose level.
    string str_vlevel = "General";
    Bridge::VerboseLevel vl = vout.set_verbose_level(str_vlevel);
    vout.general(vl, "  vlevel    = %s\n", str_vlevel.c_str());

    // gauge configuration.
    Field_G *U = new Field_G(Nvol, Ndim);

    // setup gauge configuration.
    string conf_type = "Text";
    string conf_file = "./conf_04040408.txt";
    vout.general(vl, "  conf_type = %s\n", conf_type.c_str());
    vout.general(vl, "  conf_name = %s\n", conf_file.c_str());
    GaugeConfig *gconf_read = new GaugeConfig(conf_type);
    gconf_read->read_file_4x4x4x8((Field *)U, conf_file);
    //    gconf_read->read_file((Field *)U, conf_file);

    // staple measurement for check.
    Staples *staple = new Staples;

    double plaq_expected = 0.62512994697494;
    vout.general(vl, "plaq(expected) = %18.14f\n", plaq_expected);

    double plaq = staple->plaquette(*U);
    vout.general(vl, "plaq(measured) = %18.14f\n", plaq);

    // ####  object setup  #####

    // setup gamma matrix set.
    string str_gmset_type = "Chiral";
    GammaMatrixSet *gmset = new GammaMatrixSet_Chiral();


    // setup fermion operator.
    Fopr_Wilson     *fopr_w  = new Fopr_Wilson(str_gmset_type);
    Fopr_Domainwall *fopr_dw = new Fopr_Domainwall(fopr_w);
    double mq = 0.1;
    double M0 = 1.6;
    int    Ns = 16;
    valarray<int> boundary_condition(4);
    boundary_condition[0] =  1;
    boundary_condition[1] =  1;
    boundary_condition[2] =  1;
    boundary_condition[3] = -1;
    fopr_dw->set_parameters(mq, M0, Ns, boundary_condition);
    fopr_dw->set_config(U);

    // setup sover.
    Solver *solver = new Solver_CG(fopr_dw);
    int Niter = 4000;
    double Stop_cond = 1.0e-28;
    solver->set_parameters(Niter, Stop_cond);
    // If you need to observe each step of iterations, remove the
    // commont-out of the next line.
    // solver->set_parameter_verboselevel(Bridge::DETAILED);

    // setup propagator solver.
    Fprop_Domainwall_4d *fprop = new Fprop_Domainwall_4d(fopr_dw, solver);

    Source_Local *source = new Source_Local();
    valarray<int> source_position(Ndim);
    for(int i = 0; i < Ndim; ++i){
      source_position[i] = 0;
    }
    source->set_parameters(source_position);

    Timer *timer = new Timer;

    // ####  Execution main part  ####

    std::valarray<Field_F> sq(Nc * Nd);

    double fop = 0.0;
    timer->start();

    for (int i = 0; i < Nc * Nd; ++i) {
      sq[i] = 0.0;
    }

    Field_F b;
    b = 0.0;

    int Nconv;
    double diff;

    vout.general(vl, "\n");
    vout.general(vl, "Solving quark propagator:\n");
    vout.general(vl, "  color spin   Nconv      diff     floating point op.\n");

    for (int ispin = 0; ispin < Nd; ++ispin) {
      for (int icolor = 0; icolor < Nc; ++icolor) {
        int idx = icolor + Nc * ispin;
        source->set(b, idx);

        fprop->invert_D(sq[idx], b, Nconv, diff);

        double fop1 = (264.0 + (6720.0 + 3288.0 * Nconv) * Ns) * Lvol;
        fop += fop1;
        vout.general(vl, "   %2d   %2d   %6d  %12.4e    %12.6e\n",
                     icolor, ispin, Nconv, diff,fop1);
      }
    }
    double elapsed_time = timer->report();
    double flops  = fop/elapsed_time;
    double gflops = flops * 1.0e-9;

    // hadron correlators
    vout.general(vl, "\n");
    vout.general(vl, "2-point correlator:\n");
    Corr2pt_4spinor corr(gmset);
    double result = corr.meson_all(sq, sq);

    // tidy up.
    delete solver;
    delete source;
    delete fopr_dw;
    delete fopr_w;
    delete gmset;
    delete staple;
    delete gconf_read;
    delete U;
    delete timer;

    // check the result.
    int Lx = CommonParameters::Lx();
    int Ly = CommonParameters::Ly();
    int Lz = CommonParameters::Lz();
    int Lt = CommonParameters::Lt();
    double expected_result = 0.0;
    // several examples:
    if( (Lx== 4) && (Ly== 4) && (Lz== 4) && (Lt== 8) && (Ns== 8) ){
      expected_result = 0.434204939377926;
    }
    if( (Lx== 8) && (Ly== 8) && (Lz== 8) && (Lt==16) && (Ns== 8) ){
      expected_result = 0.42434505733359;
    }
    if( (Lx== 8) && (Ly== 8) && (Lz== 8) && (Lt==16) && (Ns==16) ){
      expected_result = 0.42457877121679;
    }
    if( (Lx==16) && (Ly==16) && (Lz==16) && (Lt==32) && (Ns==16) ){
      expected_result = 0.42358228036777;
    }
    if( (Lx==32) && (Ly==32) && (Lz==32) && (Lt==32) && (Ns==16) ){
      expected_result = 0.42365191680334;
    }

    vout.general(vl, "expected = %18.14f\n", expected_result);
    vout.general(vl, "result   = %18.14f\n", result);

    if(fabs(expected_result - result) < 1.0e-11){
      vout.crucial(vl, "test result: correct! (^_^)\n");
    }else{
      vout.crucial(vl, "test result: incorrect! (T_T)\n");
    }

    vout.general(vl, "\n");
    vout.general(vl, "Performance of Domain-wall solver:\n");
    vout.general(vl, "  Floating point op. = %12.6e\n", fop);
    vout.general(vl, "  Elapsed time = %15.6f sec\n", elapsed_time);
    vout.general(vl, "  Performance  = %12.3f GFlops\n", gflops);
    vout.general(vl, "\n");

    return 0;

  }
} // namespace Test_Spectrum_Domainwall
